# Copyright (c) 2017 Ruud de Jong
# This file is part of the SlipLib project which is released under the MIT license.
# See https://github.com/rhjdjong/SlipLib for details.


"""
The :mod:`sliplib` module implements the encoding and decoding
functionality for SLIP packets, as described in :rfc:`1055`.
It defines encoding, decoding, and validation functions,
as well as various classes that can be used to to wrap
the SLIP protocol over different kinds of byte streams.

The SLIP protocol is described in :rfc:`1055` (:title:`A Nonstandard for
Transmission of IP Datagrams over Serial Lines: SLIP`, J. Romkey,
June 1988).  The original purpose of the protocol is
to provide a mechanism to indicate the boundaries of IP packets,
in particular when the IP packets are sent over a connection that
does not provide a framing mechanism, such as serial lines or
dial-up connections.

There is, however, nothing specific to IP in the SLIP protocol.
The protocol describes a generic framing method that can be used for any
type of data that must be transmitted over a (continuous) byte stream.
In fact, the main reason for creating this module
was the need to communicate with a third-party application that
used SLIP over TCP (which is a continuous byte stream)
to frame variable length data structures.

The SLIP protocol uses four special byte values:

=============== ================ =============================================
Byte value      Name             Purpose
=============== ================ =============================================
:code:`0xc0`    :const:`END`     to delimit messages
:code:`0xdb`    :const:`ESC`     to escape :data:`END` or :data:`ESC` bytes in the message
:code:`0xdc`    :const:`ESC_END` the escaped value of the :data:`END` byte
:code:`0xdd`    :const:`ESC_ESC` the escaped value of the :data:`ESC` byte
=============== ================ =============================================

An :const:`END` byte in the message is encoded as the sequence
:code:`ESC+ESC_END` (:code:`b'\\xdb\\xdc'`)
in the slip packet,
and an :const:`ESC` byte  in the message is encoded
as the sequence :code:`ESC+ESC_ESC` (:code:`b'\\xdb\\xdd'`).

.. csv-table::
   :header: "Decoded", "Encoded"

   :code:`b'\\xc0'`, :code:`b'\\xdb\\xdc'`
   :code:`b'\\xdb'`, :code:`b'\\xdb\\xdd'`

As a consequence, an :const:`ESC` byte in an encoded SLIP packet
must always be followed by an :const:`ESC_END` or an :const:`ESC_ESC` byte;
anything else is a protocol error.
"""

from .slip import *
from .sliprequesthandler import *
from .slipsocket import *
from .slipstream import *
from .version import __version__

__all__ = ['encode',
           'decode',
           'is_valid',
           'Driver',
           'SlipWrapper',
           'SlipSocket',
           'SlipRequestHandler',
           'SlipStream',
           'ProtocolError',
           'END',
           'ESC',
           'ESC_END',
           'ESC_ESC',
           '__version__']



# Copyright (c) 2017 Ruud de Jong
# This file is part of the SlipLib project which is released under the MIT license.
# See https://github.com/rhjdjong/SlipLib for details.

import collections
import sys

from .slip import Driver, ProtocolError


class SlipWrapper:
    """Base class that provides a message based interface to a byte stream

    :class:`SlipWrapper` combines a :class:`Driver` instance with a byte stream.
    The :class:`SlipWrapper` class is an abstract base class.
    It offers the methods :meth:`send_msg` and :meth:`recv_msg` to send and
    receive single messages over the byte stream, but it does not of itself
    provide the means to interact with the stream.

    To interact with a concrete stream, a derived class must implement
    the methods :meth:`send_bytes` and :meth:`recv_bytes`
    to write to and read from the stream.
    """

    def __init__(self, stream):
        """
        To instantiate a :class:`SlipWrapper`, the user must provide
        an existing byte stream

        :param stream: the byte stream that will be wrapped..
        """
        self.stream = stream
        self.driver = Driver()
        self._messages = collections.deque()
        self._protocol_error = None
        self._traceback = None
        self._flush_needed = False
        self._stream_closed = False

    def send_bytes(self, packet):
        """Send a packet over the stream.

        Derived classes must override this method.

        :param bytes packet: the packet to send over the stream
        """
        raise NotImplementedError

    def recv_bytes(self):
        """Receive data from the stream.

        Derived classes must override this method.

        .. note::
            The convention used within the :class:`SlipWrapper` class
            is that :meth:`recv_bytes` returns an empty bytes object
            to indicate that the end of
            the byte stream has been reached, and no further data will
            be received. Derived implementations must ensure that
            this convention is followed.

        :return: bytes received from the stream
        :rtype: bytes
        """
        raise NotImplementedError

    def send_msg(self, message):
        """Send a SLIP-encoded message over the stream.

        :param bytes message: The message to encode and send
        """
        packet = self.driver.send(message)
        self.send_bytes(packet)

    def recv_msg(self):
        """Receive a single message from the stream.

        :return: A SLIP-decoded message
        :rtype: bytes
        :raises ProtocolError: when a SLIP protocol error has been encountered.
           A subsequent call to :meth:`recv_msg` (after handling the exception)
           will return the message from the next packet.
        """

        # First check if there are any pending messages
        if self._messages:
            return self._messages.popleft()

        # No pending messages left. If a ProtocolError has occurred
        # it must be re-raised here:
        if self._protocol_error:
            self._handle_pending_protocol_error()

        while not self._messages and not self._stream_closed:
            # As long as no messages are available,
            # flush the internal packet buffer,
            # and try to read data
            try:
                if self._flush_needed:
                    self._flush_needed = False
                    self._messages.extend(self.driver.flush())
                else:
                    data = self.recv_bytes()
                    if data == b'':
                        self._stream_closed = True
                    self._messages.extend(self.driver.receive(data))
            except ProtocolError as pe:
                self._messages.extend(self.driver.messages)
                self._protocol_error = pe
                self._traceback = sys.exc_info()[2]
                break

        if self._messages:
            return self._messages.popleft()

        if self._protocol_error:
            self._handle_pending_protocol_error()
        else:
            return b''

    def _handle_pending_protocol_error(self):
        try:
            raise self._protocol_error.with_traceback(self._traceback)
        finally:
            self._protocol_error = None
            self._traceback = None
            self._flush_needed = True



# Copyright (c) 2017 Ruud de Jong
# This file is part of the SlipLib project which is released under the MIT license.
# See https://github.com/rhjdjong/SlipLib for details.

from socketserver import BaseRequestHandler

from .slipsocket import SlipSocket


class SlipRequestHandler(BaseRequestHandler):
    """Base class for request handlers for SLIP-based communication

    This class is derived from :class:`socketserver.BaseRequestHandler`
    for the purpose of creating TCP server instances
    that can handle incoming SLIP-based connections.

    This base class ensures that the connection
    is wrapped in a SlipSocket, and makes it
    available as :code:`self.request`.
    The original connection is still available as self.request.socket.
    The client address as available as :code:`self.client_address`,
    and the server as :code:`self.server`.

    To implement a specific behaviour, all that
    is needed is to derive a class that
    defines a :meth:`handle` method that uses
    :code:`self.request` to send and receive messages.
    """

    def setup(self):
        if not isinstance(self.request, SlipSocket):
            # noinspection PyTypeChecker
            self.request = SlipSocket(self.request)



# Copyright (c) 2017 Ruud de Jong
# This file is part of the SlipLib project which is released under the MIT license.
# See https://github.com/rhjdjong/SlipLib for details.

import socket

from .slipwrapper import SlipWrapper


class SlipSocket(SlipWrapper):
    """SlipSocket -> Wraps a TCP :class:`socket` with a :class:`Driver`

    :class:`SlipSocket` combines a :class:`Driver` instance with a
    :class:`socket`.
    The :class:`SlipStream` class has all the methods from its base class :class:`SlipWrapper`.
    In addition it directly exposes all methods and attributes of
    the contained :obj:`socket`, except for the following:

    * :meth:`send*` and :meth:`recv*`. These methods are not
      supported, because byte-oriented send and receive operations
      would invalidate the internal state maintained by :class:`SlipSocket`.
    * Similarly, :meth:`makefile` is not supported, because byte- or line-oriented
      read and write operations would invalidate the internal state.
    * :meth:`share` (Windows only) and :meth:`dup`. The internal state of
      the :class:`SlipSocket` would have to be duplicated and shared to make these methods meaningful.
      Because of the lack of a convincing use case for this, sharing and duplication is
      not supported.
    * The :meth:`accept` method is delegated to the contained :class:`socket`,
      but the socket that is returned by the :class:`socket`'s :meth:`accept` method
      is automatically wrapped in a :class:`SlipSocket` object.

    In stead of the :class:`socket`'s :meth:`send*` and :meth:`recv*` methods
    a :class:`SlipSocket` provides the method :meth:`send_msg` and :meth:`recv_msg`
    to send and receive SLIP-encoded messages.

    Only TCP sockets are supported. Using the SLIP protocol on
    UDP sockets is not supported for the following reasons:

    * UDP is datagram-based. Using SLIP with UDP therefore
      introduces ambiguity: should SLIP packets be allowed to span
      multiple UDP datagrams or not?
    * UDP does not guarantee delivery, and does not guarantee that
      datagrams are delivered in the correct order.

    """
    _chunk_size = 4096

    def __init__(self, sock):
        """
        To instantiate a :class:`SlipSocket`, the user must provide
        a pre-constructed TCP :class:`socket`.
        An alternative way to instantiate s SlipSocket is to use the
        class method :meth:`create_connection`.

        :param socket.socket sock: an existing TCP socket, i.e.
           a socket with type :const:`socket.SOCK_STREAM`
        """

        if not isinstance(sock, socket.socket) or sock.type != socket.SOCK_STREAM:
            raise ValueError('Only sockets with type SOCK_STREAM are supported')
        super().__init__(sock)
        self.socket = self.stream

    def send_bytes(self, packet):
        self.socket.sendall(packet)

    def recv_bytes(self):
        return self.socket.recv(self._chunk_size)

    def accept(self):
        conn, address = self.socket.accept()
        return self.__class__(conn), address

    @property
    def family(self):
        return self.socket.family

    @property
    def type(self):
        return self.socket.type

    @property
    def proto(self):
        return self.socket.proto

    def __getattr__(self, attribute):
        if attribute.startswith('recv') or attribute.startswith('send') or attribute in (
                'makefile', 'share', 'dup',
        ):
            raise AttributeError("'{}' object has no attribute '{}'".
                                 format(self.__class__.__name__, attribute))
        return getattr(self.socket, attribute)

    @classmethod
    def create_connection(cls, address, timeout=None, source_address=None):
        """Create a SlipSocket connection.

        This convenience method creates a connection to the the specified address
        using the :func:`socket.create_connection` function.
        The socket that is returned from that call is automatically wrapped in
        a :class:`SlipSocket` object.

        .. note::
            The :meth:`create_connection` method does not magically turn the
            socket at the remote address into a SlipSocket.
            For the connection to work properly,
            the remote socket must already
            have been configured to use the SLIP protocol.
        """
        sock = socket.create_connection(address, timeout, source_address)
        return cls(sock)



# Copyright (c) 2017 Ruud de Jong
# This file is part of the SlipLib project which is released under the MIT license.
# See https://github.com/rhjdjong/SlipLib for details.

import collections
import re

END = b'\xc0'
ESC = b'\xdb'
ESC_END = b'\xdc'
ESC_ESC = b'\xdd'
"""These constants represent the special SLIP bytes"""


class ProtocolError(ValueError):
    """Exception to indicate that a SLIP protocol error has occurred.

    This exception is raised when an attempt is made to decode
    a packet with an invalid byte sequence.
    An invalid byte sequence is either an :const:`ESC` byte followed
    by any byte that is not an :const:`ESC_ESC` or :const:`ESC_END` byte,
    or a trailing :const:`ESC` byte as last byte of the packet.

    The :exc:`ProtocolError` carries the invalid packet
    as the first (and only) element in in its :attr:`args` tuple.
    """


def encode(msg):
    """encode(msg) -> SLIP-encoded message.

    Encodes a message (a byte sequence) into a
    SLIP-encoded packet.

    :param bytes msg: The message that must be encoded
    :return: The SLIP-encoded message
    :rtype: bytes
    """
    msg = bytes(msg)
    return END + msg.replace(ESC, ESC + ESC_ESC).replace(END, ESC + ESC_END) + END


def decode(packet):
    """decode(packet) -> message from SLIP-encoded packet

    Retrieves the message from the SLIP-encoded packet.

    :param bytes packet: The SLIP-encoded message.
           Note that this must be exactly one complete packet.
           The :func:`decode` function does not provide any buffering
           for incomplete packages, nor does it provide support
           for decoding data with multiple packets.
    :return: The decoded message
    :rtype: bytes
    :raises ProtocolError: if the packet contains an invalid byte sequence.
    """
    packet = bytes(packet).strip(END)
    if not is_valid(packet):
        raise ProtocolError(packet)
    return packet.strip(END).replace(ESC + ESC_END, END).replace(ESC + ESC_ESC, ESC)


def is_valid(packet):
    """is_valid(packet) -> indicates if the packet's contents conform to the SLIP specification.
    
    A packet is valid if:

    * It contains no :const:`END` bytes other than leading and/or trailing :const:`END` bytes, and
    * Each :const:`ESC` byte is followed by either an :const:`ESC_END` or an :const:`ESC_ESC` byte.

    :param bytes packet: The packet to inspect.
    :return: :const:`True` if the packet is valid, :const:`False` otherwise
    :rtype: bool
    """
    packet = packet.strip(END)
    return not (END in packet or
                packet.endswith(ESC) or
                re.search(ESC + b'[^' + ESC_END + ESC_ESC + b']', packet))


class Driver:
    """Class to handle the SLIP-encoding and decoding of messages

    This class manages the handling of encoding and decoding of
    messages according to the SLIP protocol.
    """

    def __init__(self):
        self._recv_buffer = b''
        self._packets = collections.deque()
        self._messages = []

    # noinspection PyMethodMayBeStatic
    def send(self, message):
        """send(message) -> SLIP-encoded message.

        Encodes a message (any arbitrary byte-string) into a
        SLIP-encoded packet.

        :param bytes message: The message that must be encoded.
        :return: A packet with the SLIP-encoded message.
        :rtype: bytes
        """
        return encode(message)

    def receive(self, data):
        """receive(data) -> List of decoded messages.

        Processes :obj:`data`, which must be a bytes-like object,
        and returns a (possibly empty) list with :class:`bytes` objects,
        each containing a decoded message.
        Any non-terminated SLIP packets in :obj:`data`
        are buffered, and processed with the next call to :meth:`receive`.

        :param bytes data: The bytes-like object to be processed.
            An empty :obj:`data` parameter forces the internal
            buffer to be flushed and decoded.
        :return: A (possibly empty) list of decoded messages.
        :rtype: list(bytes)
        :raises ProtocolError: An invalid byte sequence has been detected.
        """

        # Empty data indicates that the data reception is complete.
        # To force a buffer flush, an END byte is added, so that the
        # current contents of _recv_buffer will form a complete message.
        if not data:
            data = END

        self._recv_buffer += data

        # The following situations can occur:
        #
        #  1) _recv_buffer is empty or contains only END bytes --> no packets available
        #  2) _recv_buffer contains non-END bytes -->  packets are available
        #
        # Strip leading END bytes from _recv_buffer to avoid handling empty _packets.
        self._recv_buffer = self._recv_buffer.lstrip(END)
        if self._recv_buffer:
            # The _recv_buffer contains non-END bytes.
            # It is now split on sequences of one or more END bytes.
            # The trailing element from the split operation is a possibly incomplete
            # packet; this element is therefore used as the new _recv_buffer.
            # If _recv_buffer contains one or more trailing END bytes,
            # (meaning that there are no incomplete packets), then the last element,
            # and therefore the new _recv_buffer, is an empty bytes object.
            self._packets.extend(re.split(END + b'+', self._recv_buffer))
            self._recv_buffer = self._packets.pop()

        # Process the buffered packets
        return self.flush()

    def flush(self):
        """flush() -> List of decoded messages.

        Decodes the packets in the internal buffer.
        This enables the continuation of the processing
        of received packets after a :exc:`ProtocolError`
        has been handled.

        :return: A (possibly empty) list of decoded messages from the buffered packets.
        :rtype: list(bytes)
        :raises ProtocolError: An invalid byte sequence has been detected.
        """
        messages = []
        while self._packets:
            p = self._packets.popleft()
            try:
                msg = decode(p)
            except ProtocolError:
                # Add any already decoded messages to the exception
                self._messages = messages
                raise
            messages.append(msg)
        return messages

    @property
    def messages(self):
        """messages -> List of decoded messages

        The read-only attribute :attr:`messages` contains
        the messages that were
        already decoded before a
        :exc:`ProtocolError` was raised.
        This enables the handler of the :exc:`ProtocolError`
        exception to recover the messages up to the
        point where the error occurred.
        This attribute is cleared after it has been read.
        """
        try:
            return self._messages
        finally:
            self._messages = []



# Copyright (c) 2017 Ruud de Jong
# This file is part of the SlipLib project which is released under the MIT license.
# See https://github.com/rhjdjong/SlipLib for details.

import io

from .slipwrapper import SlipWrapper


class SlipStream(SlipWrapper):
    """Class that wraps an IO stream with a :class:`Driver`

    :class:`SlipStream` combines a :class:`Driver` instance with a
    :class:`BufferedIOBase` stream.

    The :class:`SlipStream` class has all the methods and attributes
    from its base class :class:`SlipWrapper`.
    In addition it directly exposes all methods and attributes of
    the contained :obj:`stream`, except for the following:

    * :meth:`read*` and :meth:`write*`. These methods are not
      supported, because byte-oriented read and write operations
      would invalidate the internal state maintained by :class:`SlipStream`.
    * Similarly, :meth:`seek`, :meth:`tell`, and :meth:`truncate` are not supported,
      because repositioning or truncating the stream would invalidate the internal state.
    * :meth:`raw`, :meth:`detach` and other methods that provide access to or manipulate
      the stream's internal data.

    In stead of the :meth:`read*` and :meth:`write*` methods
    a :class:`SlipSocket` provides the method :meth:`send_msg` and :meth:`recv_msg`
    to send and receive SLIP-encoded messages.
    """
    _chunk_size = io.DEFAULT_BUFFER_SIZE

    def __init__(self, stream):
        """
        To instantiate a :class:`SlipStream`, the user must provide
        a pre-constructed :class:`BufferedIOBase` stream.

        :param io.BufferedIOBase stream: an existing BufferedIOBase stream.
        """
        if not isinstance(stream, io.BufferedIOBase):
            raise ValueError('Only BufferedIOBase streams are supported')
        super().__init__(stream)

    def send_bytes(self, packet):
        self.stream.write(packet)

    def recv_bytes(self):
        return b'' if self.stream.closed else self.stream.read(self._chunk_size)

    def __getattr__(self, attribute):
        if attribute in (
                'detach', 'getbuffer', 'getvalue', 'peek', 'raw',
                'read', 'read1', 'readinto', 'readinto1', 'readline', 'readlines',
                'seek', 'tell', 'truncate', 'write', 'writelines'
        ):
            raise AttributeError("'{}' object has no attribute '{}'".
                                 format(self.__class__.__name__, attribute))
        return getattr(self.stream, attribute)



# Copyright (c) 2017 Ruud de Jong
# This file is part of the SlipLib project which is released under the MIT license.
# See https://github.com/rhjdjong/SlipLib for details.

from setuptools import setup, find_packages  # Always prefer setuptools over distutils
import os.path


version_dict = {}
with open(os.path.join('src', 'sliplib', 'version.py')) as version_file:
    exec(version_file.read(), version_dict)
__version__ = version_dict['__version__']
del version_dict


# Get the long description from the relevant file
def read_long_description(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    seperator = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return seperator.join(buf)


long_description = read_long_description('README.rst')

TEST_REQUIRES = ['pytest', 'pytest-mock', 'pytest-cov']
DOC_REQUIRES = ['sphinx_rtd_theme']
DEV_REQUIRES = ['tox', 'mypy', 'wheel', 'twine']

setup(
    name='sliplib',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=__version__,

    description='Slip package',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/rhjdjong/SlipLib',

    # Author details
    author='Ruud de Jong',
    author_email='ruud.de.jong@xs4all.nl',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',

        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    # What does your project relate to?
    keywords='slip message framing protocol RFC1055',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    package_dir={'': 'src'},
    packages=find_packages('src'),

    extras_require={
        'dev': DEV_REQUIRES + DOC_REQUIRES + TEST_REQUIRES,
        'test': TEST_REQUIRES
    }
)
